/*********************************************************************
 * Software License Agreement (BSD License)
 *
 *  Copyright (c) 2022, Kassow Robots
 *  All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *   * Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *   * Redistributions in binary form must reproduce the above
 *     copyright notice, this list of conditions and the following
 *     disclaimer in the documentation and/or other materials provided
 *     with the distribution.
 *   * Neither the name of the Kassow Robots nor the names of its
 *     contributors may be used to endorse or promote products derived
 *     from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 *  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 *  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 *  FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 *  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 *  BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 *  LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 *  CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 *  LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
 *  ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 *  POSSIBILITY OF SUCH DAMAGE.
 *********************************************************************/

#ifndef KR2_KORD_UTILS_H
#define KR2_KORD_UTILS_H

#include "kr2_rc_api/internals/console.h"

#include <stdio.h>
#include <pthread.h>
#include <iostream>
#include <sys/mman.h>

//#include <fstream>

#include <stdint.h>

#include <fcntl.h>

// following are all for the stat function
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <math.h>

#include <signal.h>

#include <chrono>

namespace kr2 {
namespace utils {

static const int DEFAULT_PRIORITY = 50;
static const int DEFAULT_PRIORITY_MAX = 50;

class Stats {

public:
    using timespan = std::chrono::duration<int64_t, std::nano>;

    Stats() {
        reset();
    }

    void reset() {
        using namespace std::chrono_literals;
        min_ = 1000s;
        max_ = 0us;
    }

    void updateFailure(const std::chrono::time_point<std::chrono::system_clock> &a_capture_ts){
        failed_capture_cnt_++;
    }

    void updateCapture(const std::chrono::time_point<std::chrono::system_clock> &a_capture_ts){
        if (captured_message_cnt_ == 0){
            cap_first_ = a_capture_ts;
            cap_latest_ = a_capture_ts;
            captured_message_cnt_++;
            return;
        }

        timespan elapsed = a_capture_ts - cap_latest_;
        if (elapsed < min_){
            min_ = elapsed;
        }

        if (elapsed > max_){
            max_ = elapsed;
        }

        cap_latest_ = a_capture_ts;
        captured_message_cnt_++;
    }

    void printStats(std::ostream & a_stream) const{
        int64_t avg_cap = std::chrono::duration_cast<timespan>(cap_latest_ - cap_first_).count();
        avg_cap /= static_cast<int64_t>(captured_message_cnt_);
        a_stream << "Captured msgs; Failed msgs; min[ns]; avg capture[ns]; max[ns]\n";
        a_stream << captured_message_cnt_ << ";  " << failed_capture_cnt_ << "; " << min_.count() << "; " << avg_cap << "; " << max_.count() << "\n";
    }

    int64_t get_average(){
        int64_t avg_cap = std::chrono::duration_cast<timespan>(cap_latest_ - cap_first_).count();
        return avg_cap / static_cast<int64_t>(captured_message_cnt_);
    }

    int64_t get_min(){
        return min_.count();
    }

    int64_t get_max(){
        return max_.count();
    }

private:
    unsigned long captured_message_cnt_ = 0;
    unsigned int failed_capture_cnt_ = 0;

    std::chrono::time_point<std::chrono::system_clock> cap_first_;
    std::chrono::time_point<std::chrono::system_clock> cap_latest_;
    timespan min_;
    timespan max_;
};

struct StampsLog {
    int64_t tick_;
    int64_t t0_;
    int64_t t1_;
    int64_t t2_;

    void reset() {
        tick_ = INT64_MAX;
        t0_ = INT64_MAX;
        t1_ = INT64_MAX;
        t2_ = INT64_MAX;
    }

    std::string header(){
        std::stringstream ss;
        ss << "tick, t0, t1, t2";
        return ss.str();
    }

    std::string asString(){
        std::stringstream ss;
        ss << tick_ << ", " << t0_ << ", " << t1_ << ", " << t2_;
        return ss.str();
    }
};


bool set_latency_target(void)
{
    struct stat s;
    int latency_target_fd = -1;
    int err;

    err = stat("/dev/cpu_dma_latency", &s);
    if (err == -1) {
        CLOG_INFO("[RealTime] getting file '/dev/cpu_dma_latency' status failed");
        return false;
    }

    latency_target_fd = open("/dev/cpu_dma_latency", O_RDWR);
    if (latency_target_fd == -1) {
        CLOG_INFO("[RealTime] opening '/dev/cpu_dma_latency' failed");
        return false;
    }

    int32_t latency_target_value = 0;

    ssize_t written = write(latency_target_fd, &latency_target_value, 4);
    if (written < 1) {
        CLOG_INFO("[RealTime] setting the cpu_dma_latency to " << latency_target_value << "us failed");
        close(latency_target_fd);
        return false;
    }
    CLOG_INFO("[RealTime] /dev/cpu_dma_latency set to " << latency_target_value << "us");

    close(latency_target_fd);

    return true;
}

void init_realtime_params(int a_prio)
{
    bool rt_complete = true;

    try {
        CLOG_INFO("[RealTime] setting DMA latency target");
        if(!set_latency_target()) {
            CLOG_INFO("[RealTime] setting DMA latency target failed");
            throw std::exception();
        }

        // We'll operate on the currently running thread.
        pthread_t this_thread = pthread_self();

        // struct sched_param is used to store the scheduling priority
        struct sched_param params;
        // We'll set the priority to the maximum.
        params.sched_priority = DEFAULT_PRIORITY;

        if ((a_prio > 0) && (a_prio <= DEFAULT_PRIORITY_MAX)) {
            params.sched_priority = a_prio;
        }

        // Attempt to set thread real-time priority to the SCHED_FIFO policy
        CLOG_INFO("[RealTime] setting thread realtime priority = " << params.sched_priority);
        int ret_value = pthread_setschedparam(this_thread, SCHED_FIFO, &params);
        if (ret_value != 0) {
            // Print the error
            CLOG_INFO("[RealTime] setting thread real-time priority failed");
            throw std::exception();
        }
        // Now verify the change in thread priority
        int policy = 0;
        ret_value = pthread_getschedparam(this_thread, &policy, &params);
        if (ret_value != 0) {
            CLOG_INFO("[RealTime] retrieving real-time scheduling parameters failed");
            throw std::exception();
        }

        // Check the correct policy was applied
        if(policy != SCHED_FIFO) {
            CLOG_INFO("Scheduling is NOT SCHED_FIFO!" << std::endl);
            CLOG_INFO("[RealTime] 'SCHED_FIFO' scheduling not okay");
            throw std::exception();
        } else {
            CLOG_INFO("[RealTime] 'SCHED_FIFO' ok");
        }

        // Print thread scheduling priority
        CLOG_INFO("[RealTime] thread priority is: " << params.sched_priority);

        // block SIGALRM
        sigset_t sigset;
        sigemptyset(&sigset);
        sigaddset(&sigset, SIGALRM);
        sigprocmask(SIG_BLOCK, &sigset, NULL);

        // Lock process pages in RAM from swapping
        CLOG_INFO("[RealTime] memory lock all");
        if (mlockall(MCL_FUTURE))
        {
            CLOG_INFO("[RealTime] mlockall failed");
            throw std::exception();
        } else {
            CLOG_INFO("[RealTime] mlockall successful");
        }

        // Try detaching this thread
        pthread_detach(this_thread);
    } catch (...) {
         rt_complete = false;
    }
    if (rt_complete) {
        CLOG_INFO("[RealTime] real-time parameters successfuly set\n");
    } else {
        CLOG_INFO("[RealTime] real-time settings failed!\n");
    }

}

int64_t timespecToEpochNanosec(const timespec & ts){
    return std::chrono::high_resolution_clock::time_point(
            std::chrono::seconds(ts.tv_sec)
            +std::chrono::nanoseconds(ts.tv_nsec))
            .time_since_epoch()
            .count();
}

} // namespace kord
} // namespace kr2

#endif // KR2_KORD_UTILS_H
